from functools import partial
from multiprocessing import Pool
import os

import numpy as np                     # Pairwise distances.
from expam.sequences import get_sequence

try:
    from sourmash import signature as sm_signature
    import sourmash

except ModuleNotFoundError:
    print("Could not import sourmash! Install sourmash to use this feature.")


def make_signature(genome_url, k, s):
    print("Generating signature for %s." % genome_url)

    minhash = sourmash.MinHash(n=0, ksize=k, scaled=s)
    sequence = get_sequence(genome_url)
    minhash.add_sequence(sequence, True)

    sig = sm_signature.SourmashSignature(
        minhash, name=os.path.basename(genome_url), filename=genome_url)

    return sig


def make_signatures(n_processes, genome_dirs, sig_dir, k, s):
    print("Creating signatures!")

    if not os.path.exists(sig_dir):
        os.makedirs(sig_dir)

        print("Created directory %s." % sig_dir)

    else:
        # Check for any genomes already processed.
        sig_names = [
            os.path.basename(f).replace(".sour", "")
            for f in os.listdir(sig_dir)
        ]

        genome_dirs = [
            genome_dir
            for genome_dir in genome_dirs
            if os.path.basename(genome_dir) not in sig_names
        ]

    # Calculate chunksize.
    chunksize = 100 if (len(genome_dirs) // n_processes) > 100 else (len(genome_dirs) // n_processes)

    print("Starting process pool...")

    with Pool(n_processes) as mp_pool:
        sigs = list(mp_pool.imap_unordered(
            partial(make_signature, k=k, s=s),
            genome_dirs,
            chunksize=chunksize
        ))

    print("Process pool completed!")

    # Save sigs to disk.
    sig_data = sm_signature.save_signatures(sigs).decode('utf-8')

    with open(os.path.join(sig_dir, "mysigs.sig"), "w") as f:
        f.write(sig_data)


def make_distances(sig_dir, n_processes, matrix_dir):
    if not os.path.exists(sig_dir):
        raise Exception("No such signatures at %s!" % sig_dir)

    with open(os.path.join(sig_dir, "mysigs.sig"), "r") as f:
        sig_json = f.read()

    signatures = list(sm_signature.load_signatures(sig_json))
    distance_matrix = sourmash.compare.compare_all_pairs(
        siglist=signatures,
        ignore_abundance=True,
        downsample=False,
        n_jobs=n_processes
    )

    np.fill_diagonal(distance_matrix, 0)
    names = [
        sig.name[:-4] if sig.name[-4:] == ".fna" else sig.name
        for sig in signatures
    ]
    to_phylip_matrix(distance_matrix, names, matrix_dir)


def to_phylip_matrix(distance_matrix, names, out_dir):
    phylip_string = "%d\n" % len(distance_matrix)

    for i, row in enumerate(distance_matrix):
        phylip_string += "%s\t%s\n" % (
            names[i],
            "\t".join(
                map(str, row)
            )
        )

    with open(out_dir, "w") as f:
        f.write(phylip_string)



#!/usr/bin/env python3

import logging
import multiprocessing as mp
import psutil
import re
import resource
import subprocess
import time


def limit(mem_limit):
    """
    Defines a wrapper for functions that first sets a resource limit before
    executing the function.

    :param mem_limit:
    :return:
    """
    def inner(fn):
        """
        Wrapper that sets some memory limit then executes the input function.

        :param fn:
        :return:
        """
        def limited_fn(*args, **kwargs):
            soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
            resource.setrlimit(resource.RLIMIT_RSS, (mem_limit, hard))

            fn(*args, **kwargs)

        return limited_fn
    return inner


def sandbox(logging_name, mem_limit, interval):
    """
    Create a decorator that runs function in a new process, limiting the resources of that
    process and observing it at regular intervals.

    :param mem_limit:
    :param interval:
    :return:
    """
    def inner(fn):
        """
        Run function in a new process, modifying the function so that it first sets a resource
        limit for the environment it is in.

        :param fn:
        :return:
        """
        def job(*args, **kwargs):
            # Configure logging output.
            logger = logging.getLogger(logging_name)
            logger.info("\ttotal\tused\tfree\tshared\tbuff/cache\tavailable")

            # Start new child process.
            ctx = mp.get_context('fork')

            new_target = limit(mem_limit)(fn)
            p = ctx.Process(target=new_target, args=args, kwargs=kwargs)

            p.start()

            while p.is_alive():
                # Log memory usage.
                mem_info = subprocess.run(['free', '-h'], capture_output=True).stdout.decode()
                logger.info("\t".join(re.findall(r"[^\s]+", mem_info)[6:13]))

                time.sleep(interval)

            p.join()

        return job
    return inner


def parse(cmd, mem_limit=None, interval=1, out=None):
    i = 0

    if len(cmd) > 0 and 'sandbox' in cmd[i]:
        i += 1

    while i < len(cmd):
        if cmd[i] in {'--memory', '-m'}:
            mem_limit = int(cmd[i + 1])
            i += 1

        elif cmd[i] in {'--interval', '-t'}:
            interval = float(cmd[i + 1])
            i += 1

        elif cmd[i] in {'--out', '-o'}:
            out = cmd[i + 1]
            i += 1

        elif cmd[i] in {'--x', '-x'}:
            total_memory = psutil.virtual_memory().total
            mem_limit = int(float(cmd[i + 1]) * total_memory)
            i += 1

        else:
            break

        i += 1

    cmd = cmd[i:]
    return mem_limit, interval, out, cmd


def sandbox_command(data):
    mem_limit, interval, out_file, cmd = parse(data)

    if mem_limit is None:
        soft, hard = resource.getrlimit(resource.RLIMIT_RSS)
        mem_limit = soft

    logger = logging.getLogger('sandbox_application')
    logger.setLevel(logging.DEBUG)

    # Create the appropriate handler.
    if out_file is None:
        handle = logging.StreamHandler()
    else:
        handle = logging.FileHandler(out_file)

    handle.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(asctime)s ... %(message)s')
    handle.setFormatter(formatter)

    logger.addHandler(handle)

    @sandbox(logging_name='sandbox_application', mem_limit=mem_limit, interval=interval)
    def run_cmd(command):
        subprocess.run(command, shell=False, check=True)

    if cmd:
        run_cmd(cmd)


def main():
    import sys
    sandbox_command(sys.argv)


if __name__ == '__main__':
    main()



import datetime
import os
import re
import requests
import shutil
import time
from ctypes import *
from math import floor
from multiprocessing import Array, shared_memory, Value
from os.path import join

import numpy as np
import pandas as pd
from expam.processes import ReadWorker, ClassifyWorker
from expam.run import EXPAM_TEMP_EXT, TIMEOUT, UNION_RATIO, ControlCenter
from expam.sequences import ls
from expam.stores import TablesDb
from expam.tree import Index

#
# Constants.
#

from expam.run import KEYS_DATA_TYPE, \
    KEYS_DATA_TYPE_SIZE, \
    MAP_DATA_TYPE, \
    MAP_DATA_TYPE_SIZE, \
    LCA_MATRIX_PATH, \
    DB_PATH, \
    ACCESSION_ID_PATH

ENTREZ_EMAIL = "ssol0002@student.monash.edu"
ENTREZ_TOOL = "expam"
ENTREZ_API = "8e5edfa1413576bca4f48b4a5520f6295308"
_STEP = 500  # NCBI Database searches at a time.

PHY_RESULTS = "phy"
TAX_RESULTS = "tax"
RAW_RESULTS = "raw"
TEMP_RESULTS = "temp"

CLASSIFICATIONS_NAME = "classifications.csv"
SPLITS_NAME = "splits.csv"
FINAL_NAME = "results.csv"

TAXID_LINEAGE_MAP_NAME = os.path.join("phylogeny", "taxid_lineage.csv")
TAXON_RANK_MAP_NAME = os.path.join("phylogeny", "taxa_rank.csv")

RANK_ORDER = ["0", "root", "top", "superkingdom", "kingdom", "subkingdom", "superclade", "clade", "subclade",
              "superphylum", "phylum", "subphylum", "superclass", "class", "subclass", "superorder", "order",
              "suborder", "superfamily", "family", "subfamily", "supergenus", "genus", "subgenus", "species group",
              "species subgroup", "superspecies", "species", "subspecies", "serogroup", "serotype", "superstrain",
              "strain", "substrain", "no rank"]
TAX_RESULTS_HEADER = ["taxid", "%ClassifiedCumulative", "ClassifiedCumulative", "ClassifiedCount",
                      "%SplitCumulative", "SplitCumulative", "SplitCount", "Rank", "Lineage"]

#
# Functions.
#


def prepare_results(n):
    # Create a shared array to put fully classified reads.
    classifications = Array(c_uint32, n, lock=True)
    splits = Array(c_uint32, n, lock=True)
    unclassified = Value(c_uint32, lock=True)

    return classifications, splits, unclassified


def name_to_id(phylogeny_path):
    # Load Index from Newick file.
    with open(phylogeny_path, "r") as f:
        newick_str = f.read().strip()

    _, indexMap = Index.from_newick(newick_str)
    # Define the index dictionary and phi mapping.
    index = {}
    for i, loc in enumerate(indexMap.pool):
        # Define the name mapping.
        index[loc.name] = i

    return index, indexMap


def string_to_tuple(st):
    opener = st.split("(")[1]
    number_string = opener.split(")")[0]
    return tuple(int(i) for i in number_string.split(",") if i != '')


def run_classifier(reads_url, out_dir, k, n, phylogeny_path, keys_shape, values_shape, logging_dir,
                   taxonomy=False, cutoff=0.0, groups=None, keep_zeros=False, cpm=0.0, use_node_names=True,
                   phyla=False, name_taxa=None, colour_list=None, results_name=None, circle_scale=1.0,
                   paired_end=False, alpha=1.0):
    def format_date():
        _date = datetime.datetime.now()
        year, month, day = _date.year, _date.month, _date.day

        date_string = ""
        for part in (year, month, day):
            if len(str(part)) == 1:
                date_string += "0" + str(part)

            else:
                date_string += str(part)

        return date_string

    # Create a dedicated folder to save results.
    # Format of results folders are ./out_dir/results/YYYYMMDD_ID/...
    results_base = os.path.join(out_dir, "results")
    if not os.path.exists(results_base):
        os.mkdir(results_base)

    if results_name is None:
        folder_base = os.path.join(results_base, format_date())

        i = 1
        while True:
            candidate = folder_base + "_" + str(i)

            if not os.path.exists(candidate):
                results_path = candidate
                break

            else:
                i += 1

    else:
        results_path = os.path.join(results_base, results_name)

        if os.path.exists(results_path):
            over_write = input("Results folder %s already exists! Overwrite? (y/n) ")

            if over_write.lower() == "y":
                print("Overwriting directory...")
                shutil.rmtree(results_path)
            else:
                print("Did not overwrite. Exiting...")
                return

    os.mkdir(results_path)
    print("Results directory created at %s." % results_path)

    print("Loading the map and phylogeny.\n")
    # Load the kmer dictionary.
    db_kmers = SharedDB(out_dir, keys_shape, values_shape)

    phy_results_path = os.path.join(results_path, PHY_RESULTS)
    raw_results_path = os.path.join(phy_results_path, RAW_RESULTS)
    temp_results_path = os.path.join(results_path, TEMP_RESULTS)

    for new_path in [phy_results_path, raw_results_path, temp_results_path]:
        os.mkdir(new_path)

    try:
        # Get name mapping.
        index, phylogeny_index = name_to_id(phylogeny_path)

        # Load LCA Matrix.
        lca_matrix = np.load(join(out_dir, LCA_MATRIX_PATH))

        # Initialise the distribution.
        d = Distribution(
            k=k,
            kmer_db=db_kmers,
            index=phylogeny_index,
            lca_matrix=lca_matrix,
            url=reads_url,
            out_dir=results_path,
            logging_dir=logging_dir,
            temp_dir=temp_results_path,
            paired_end=paired_end,
            alpha=alpha
        )

        # Run the classification with n processes.
        d.run(n)

        # Load accession ids.
        with open(join(out_dir, ACCESSION_ID_PATH), 'r') as f:
            sequence_ids = [
                line.strip().split(",")
                for line in f
            ]

        # Update accession ids in map.
        for (fname, accn_id, tax_id) in sequence_ids:
            if accn_id != "None":
                phylogeny_index[fname].accession_id = accn_id

            if tax_id != "None":
                phylogeny_index[fname].taxid = tax_id

        # Output results.
        results = ClassificationResults(
            index=index,
            phylogeny_index=phylogeny_index,
            in_dir=phy_results_path,
            out_dir=results_path,
            groups=groups,
            keep_zeros=keep_zeros,
            cutoff=cutoff,
            cpm=cpm,
            use_node_names=use_node_names,
            phyla=phyla,
            name_taxa=name_taxa,
            colour_list=colour_list,
            circle_scale=circle_scale
        )

        if taxonomy:
            # Attempt to update taxon ids.
            accession_to_taxonomy(out_dir)

            tax_results_path = os.path.join(results_path, TAX_RESULTS)
            os.mkdir(tax_results_path)

            name_to_lineage, taxon_to_rank = load_taxonomy_map(out_dir)
            results.to_taxonomy(name_to_lineage, taxon_to_rank, tax_results_path)

        results.draw_results()

    finally:

        # Close shared memory.
        keys_shm = shared_memory.SharedMemory(
            name=db_kmers.keys_shm_name,
            create=False
        )
        values_shm = shared_memory.SharedMemory(
            name=db_kmers.values_shm_name,
            create=False
        )
        keys_shm.unlink()
        keys_shm.close()
        values_shm.unlink()
        values_shm.close()

        os.rmdir(temp_results_path)


def load_taxonomy_map(out_dir, convert_to_name=True):
    def yield_lines(fh, watch=False):
        for line in fh:
            line = line.strip()
            if line:
                line = line.split(",")
                yield line

    name_tax_id_url = os.path.join(out_dir, ACCESSION_ID_PATH)
    tax_id_lineage_url = os.path.join(out_dir, TAXID_LINEAGE_MAP_NAME)
    taxa_rank_url = os.path.join(out_dir, TAXON_RANK_MAP_NAME)

    # Create map from scientific name --> (taxid, rank).
    taxon_data = {}
    with open(taxa_rank_url, "r") as f:
        for data in yield_lines(f):
            taxon_data[",".join(data[0:-2])] = tuple(data[-2:])

    # Create map from tax_id --> lineage (tuple).
    tax_id_to_lineage = {}
    with open(tax_id_lineage_url, "r") as f:
        for data in yield_lines(f):
            tax_id_to_lineage[data[0]] = tuple(data[1:])

    if not convert_to_name:
        return tax_id_to_lineage, taxon_data

    # Create map from name --> lineage (tuple).
    name_to_lineage = {}
    with open(name_tax_id_url, "r") as f:
        for data in yield_lines(f):
            name_to_lineage[data[0]] = tax_id_to_lineage[data[2]]

    return name_to_lineage, taxon_data


def load_sequence_map(out_dir):
    sequence_data_url = os.path.join(out_dir, ACCESSION_ID_PATH)

    if not os.path.exists(sequence_data_url):
        return []

    with open(sequence_data_url, "r") as f:
        sequences_ids = [
            line.strip().split(",")
            for line in f
        ]

    return sequences_ids


def load_taxid_lineage_map(out_dir):
    map_url = os.path.join(out_dir, TAXID_LINEAGE_MAP_NAME)

    if not os.path.exists(map_url):
        return []

    with open(map_url, "r") as f:
        taxid_map = [
            line.strip().split(",")
            for line in f
        ]

    return taxid_map


def load_rank_map(out_dir):
    rank_url = os.path.join(out_dir, TAXON_RANK_MAP_NAME)

    name_to_rank = {}

    if not os.path.exists(rank_url):
        return name_to_rank

    with open(rank_url, "r") as f:
        for line in f:
            line = line.strip()

            try:
                name_index = line.index(",")
                name_to_rank[line[:name_index]] = line[name_index + 1:]
            except ValueError:
                continue

    return name_to_rank


def to_first_tab(string):
    name = re.findall(r'^[^\t]+(?=\t)', string)

    if not name:
        print(string)
        raise ValueError("No name found!")

    else:
        return name[0]


def accession_to_taxonomy(out_dir):
    """
    Map accession IDs to taxonomic labels.

    :sequence_ids: List - (sequence_id, accession_id, taxon_id)
    :taxon_ranks: List - (taxon_id, taxonomic ranks)
    :taxa_to_rank: Dict - taxon --> rank
    """

    def tuples_to_disk(lst):
        return "\n".join([",".join(item) for item in lst])

    def dict_to_disk(dct):
        return "\n".join([",".join((key, value)) for key, value in dct.items()])

    sequence_ids = load_sequence_map(out_dir)
    taxon_ranks = load_taxid_lineage_map(out_dir)
    taxa_to_rank = load_rank_map(out_dir)

    # Collect taxon ids for unknown organisms.
    accessions_to_be_mapped = []
    taxa_to_be_collected = []

    for (sequence_id, accession_id, taxon_id) in sequence_ids:
        if taxon_id == "None":
            accessions_to_be_mapped.append(accession_id)
        else:
            taxa_to_be_collected.append(taxon_id)

    if accessions_to_be_mapped:
        accession_to_tax = request_tax_ids("nuccore", accessions_to_be_mapped)

        print("Received %d response(s) for ESummary TaxID request!"
              % len(accession_to_tax))

        for i in range(len(sequence_ids)):
            if sequence_ids[i][1] in accession_to_tax:
                sequence_ids[i][2] = accession_to_tax[sequence_ids[i][1]]
                taxa_to_be_collected.append(sequence_ids[i][2])

    # Collect taxonomic lineages for taxa.
    current_taxa = {taxa[0] for taxa in taxon_ranks}
    taxa_to_be_collected = {  # Set so that we collect unique values.
        taxon_id
        for taxon_id in taxa_to_be_collected
        if taxon_id not in current_taxa
    }

    if taxa_to_be_collected:
        taxid_to_taxon, taxon_to_rank = request_labels("taxonomy", "xml", list(taxa_to_be_collected))

        print("Received %d response(s) for EFetch Taxon request!"
              % len(taxid_to_taxon))

        taxon_ranks.extend(taxid_to_taxon)
        taxa_to_rank.update(taxon_to_rank)

    # Save update maps to disk.
    with open(os.path.join(out_dir, ACCESSION_ID_PATH), "w") as f:
        f.write(tuples_to_disk(sequence_ids))

    with open(os.path.join(out_dir, TAXID_LINEAGE_MAP_NAME), "w") as f:
        f.write(tuples_to_disk(taxon_ranks))

    print("Taxonomic lineages written to %s!" % os.path.join(out_dir, TAXID_LINEAGE_MAP_NAME))

    with open(os.path.join(out_dir, TAXON_RANK_MAP_NAME), "w") as f:
        f.write(dict_to_disk(taxa_to_rank))

    print("Taxonomic ranks written to %s!" % os.path.join(out_dir, TAXON_RANK_MAP_NAME))


def request_tax_ids(db, id_list):
    POST_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esummary.fcgi"
    taxids, upto = {}, 0

    PARAMS = {
        "tool": ENTREZ_TOOL,
        "email": ENTREZ_EMAIL,
        "api_key": ENTREZ_API,
        "db": db,
    }

    while upto < len(id_list):
        next_requests = id_list[upto:upto + _STEP]

        print("Posting %d UIDs to NCBI Entrez %s."
              % (len(next_requests), db))

        PARAMS["id"] = ",".join(next_requests)
        esummary_request = requests.post(
            url=POST_URL,
            data=PARAMS
        )

        # Parse TaxonIDs from raw results.
        accn_id = tax_id = None

        for line in esummary_request.text.split("\n"):
            if "<DocSum>" in line:  # Start new record.
                accn_id = tax_id = None

            elif 'AccessionVersion' in line:
                accn_id = parse_id(line)

            elif 'TaxId' in line:
                tax_id = parse_tax_id(line)

            elif "</DocSum>" in line:  # Only save complete records.
                if accn_id is not None and tax_id is not None:
                    taxids[accn_id] = tax_id

        upto += _STEP

        time.sleep(1.0)  # Allow server time to breath.

    return taxids


def parse_id(string):
    new_id = re.findall(r'\<Item Name\="AccessionVersion" Type\="String"\>(.*?)\<\/Item\>', string)

    if not new_id:
        raise ValueError("No taxids found!")

    else:
        return new_id[0]


def parse_tax_id(string):
    taxids = re.findall(r'\<Item Name\="TaxId" Type\="Integer"\>(.*?)\<\/Item\>', string)

    if not taxids:
        raise ValueError("No taxids found!")

    else:
        return taxids[0]


def request_labels(db, retmode, id_list):
    POST_URL = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
    taxon_labels, ranks, upto = [], {}, 0

    PARAMS = {
        "tool": ENTREZ_TOOL,
        "email": ENTREZ_EMAIL,
        "api_key": ENTREZ_API,
        "db": db,
        "retmode": retmode,
    }

    while upto < len(id_list):
        next_requests = id_list[upto:upto + _STEP]

        print("Posting %d UIDs to NCBI Entrez %s."
              % (len(next_requests), db))

        PARAMS["id"] = ",".join(next_requests)
        efetch_request = requests.post(
            url=POST_URL,
            data=PARAMS
        )

        # Parse taxonomic labels from raw results.
        tax_id = taxa = rank = name = None
        sub_tax_id = sub_rank = sub_name = None
        collect = True

        for line in efetch_request.text.split("\n"):
            if "<Taxon>" in line and collect:  # Start new record.
                tax_id = taxa = name = rank = None

            elif "<Taxon>" in line and not collect:
                sub_tax_id = sub_rank = sub_name = None

            elif "<TaxId>" in line and collect:
                tax_id = parse_single_tax_id(line)

            elif "<TaxId>" in line and not collect:
                sub_tax_id = parse_single_tax_id(line)

            elif "<Lineage>" in line and collect:
                taxa = parse_lineage(line).replace(",", "")

            elif "<Rank>" in line and not collect:
                if sub_name == "cellular organisms":
                    sub_rank = "top"

                else:
                    sub_rank = parse_rank(line)

            elif "<Rank>" in line and collect:
                if name == "cellular organisms":
                    rank = "top"

                else:
                    rank = parse_rank(line)

            elif "<ScientificName>" in line and not collect:
                sub_name = parse_name(line)

            elif "<ScientificName>" in line and collect:
                name = parse_name(line)

            elif "<LineageEx>" in line:
                collect = False

            elif "</Taxon>" in line and not collect:
                ranks[sub_name] = sub_tax_id + "," + sub_rank

            elif "</LineageEx>" in line:
                collect = True

            elif "</Taxon>" in line and collect:
                if tax_id is not None and taxa is not None:
                    lineage = taxa.strip().split("; ")
                    if name not in lineage and name is not None:
                        lineage += [name]

                    taxon_labels.append([tax_id, ",".join(lineage)])

                ranks[name] = tax_id + ',' + rank

        upto += _STEP

        time.sleep(1.0)  # Allow server time to breath.

    return taxon_labels, ranks


def parse_single_tax_id(string):
    taxids = re.findall(r'\<TaxId\>(.*?)\<\/TaxId\>', string)

    if not taxids:
        raise ValueError("No taxids found!")

    else:
        return taxids[0]


def parse_lineage(string):
    lineage = re.findall(r'\<Lineage\>(.*?)\<\/Lineage\>', string)

    if not lineage:
        raise ValueError("No lineages found!")

    else:
        return lineage[0]


def parse_rank(string):
    rank = re.findall(r'\<Rank\>(.*?)\<\/Rank\>', string)

    if not rank:
        raise ValueError("No rank found!")

    else:
        return rank[0]


def parse_name(string):
    name = re.findall(r'\<ScientificName\>(.*?)\<\/ScientificName\>', string)

    if not name:
        raise ValueError("No rank found!")

    else:
        name = re.sub(r"[,]", "", name[0])
        return name


def combine_results(out_dir, taxa_to_rank, cutoff=0):
    final_url = os.path.join(out_dir, FINAL_NAME)

    def where_positive(lst):
        return [x for x, y in enumerate(lst) if y][0]

    def is_number(string):
        return string.lstrip("-").replace(".", "1").replace("e-", "1").isdigit()

    def classified_counter(num):
        return float(num[0])

    def unclassified_counter(lst):
        return len(lst)

    #
    #   Format: [name, rank, classified, unclassified, [...]]
    #

    classified_url = os.path.join(out_dir, TAX_RESULTS, CLASSIFICATIONS_NAME)
    splits_url = os.path.join(out_dir, TAX_RESULTS, SPLITS_NAME)

    methods = {
        "classified": {
            "counter": classified_counter,
            "index": 2,
            "url": classified_url
        },
        "splits": {
            "counter": unclassified_counter,
            "index": 3,
            "url": splits_url
        },
    }

    results = [None, None, 0, 0, []]
    active_ranks = set()

    for method in ("classified", "splits"):
        if not os.path.exists(methods[method]["url"]):
            continue

        with open(methods[method]["url"], "r") as f:
            first_line = f.readline()
            f.seek(0)

            n_cols = first_line.count(",") + 1
            data = [[] for _ in range(n_cols)]

            for line in f:
                row = line.strip().split(",")
                for i in range(len(row)):
                    if row[i] != '':
                        data[i].append(row[i])

            for entry in data:
                check_numbers = [is_number(value) for value in entry]
                count_start = where_positive(check_numbers)

                count = methods[method]["counter"](entry[count_start:])

                if count < cutoff:
                    continue

                level = results

                for taxa in entry[:count_start]:
                    if taxa == "nan":
                        break

                    rank = taxa_to_rank[taxa]

                    if rank not in RANK_ORDER:
                        raise ValueError("Unknown rank %s for taxa %s!" % (rank, taxa))

                    else:
                        active_ranks.add(rank)

                    check_children = [taxa == child[0] for child in level[4]]

                    if not any(check_children):
                        child_index = len(level[4])
                        level[4].append([taxa, rank, 0, 0, []])

                    else:
                        child_index = where_positive(check_children)

                    level[4][child_index][methods[method]["index"]] += count
                    level = level[4][child_index]

    my_rank_order, i = {}, 0
    for rank in RANK_ORDER:

        if rank in active_ranks:
            my_rank_order[rank] = i
            i += 1

    line_format = ["" for _ in range(4 * len(active_ranks))]

    rank_line = line_format.copy()
    for rank in my_rank_order:
        rank_line[4 * my_rank_order[rank]] = rank

    filler = ["", "classified", "unclassified", ""]
    header_line = line_format.copy()
    for i, section in enumerate(line_format):
        header_line[i] = filler[i % len(filler)]

    formatted_results = ",".join(rank_line) + "\n" + ",".join(header_line) + "\n"

    def format_lineage(level, line):
        my_start_ind = my_rank_order[level[1]] * 4

        line[my_start_ind] = level[0]
        line[my_start_ind + 1] = str(level[2])
        line[my_start_ind + 2] = str(level[3])

        if len(level[4]) > 0:
            lines = [line.copy(), ] + [line_format.copy() for _ in range(len(level[4]) - 1)]

            return "\n".join([format_lineage(level[4][c], lines[c]) for c in range(len(lines))])

        else:
            return ",".join(line)

    formatted_results += format_lineage(results[4][0], line_format.copy())

    with open(final_url, "w") as f:
        f.write(formatted_results)

    print("Formatted results written to %s!" % final_url)


#
# Classes.
#


class SharedDB:
    def __init__(self, out_dir, keys_shape, values_shape):
        #
        print("Preparing shared memory allocations...")

        # Prepare shm segment for keys.
        keys_shm = shared_memory.SharedMemory(
            size=np.product(keys_shape) * KEYS_DATA_TYPE_SIZE,
            create=True
        )
        keys_interface = np.ndarray(
            shape=keys_shape,
            dtype=KEYS_DATA_TYPE,
            buffer=keys_shm.buf
        )

        # Prepare shm segment for values.
        values_shm = shared_memory.SharedMemory(
            size=np.product(values_shape) * MAP_DATA_TYPE_SIZE,
            create=True
        )
        values_interface = np.ndarray(
            shape=values_shape,
            dtype=MAP_DATA_TYPE,
            buffer=values_shm.buf
        )

        db = TablesDb(join(out_dir, DB_PATH))

        #
        print("Loading database keys...")
        db.read_keys(arr=keys_interface)

        #
        print("Loading database values...")
        db.read_values(arr=values_interface)

        db.close()

        # Remember the names for attaching in future.
        self.keys_shm_name = keys_shm.name
        self.values_shm_name = values_shm.name

        # Remember array data.
        self.keys_meta = {
            "shape": keys_shape,
            "dtype": KEYS_DATA_TYPE
        }
        self.values_meta = {
            "shape": values_shape,
            "dtype": MAP_DATA_TYPE
        }

        # Close my access.
        keys_shm.close()
        values_shm.close()


class ExpamClassifierProcesses(ControlCenter):
    @classmethod
    def from_method_dict(cls, logging_dir, config):
        base_center = super(ExpamClassifierProcesses, cls).from_dict(logging_dir, config)

        base_arguments = {
            "group_name": "group_name",
            "workers": "workers",
            "child_statuses": "_child_statuses",
            "phases": "phases",
            "phase_queues": "phase_queues",
            "child_queues": "child_queues",
            "children": "_children",
            "processors": "_processors",
            "transitions": "_transitions",
            "timeout": "_timeout",
        }
        base_attributes = {
            attr: getattr(base_center, attr_reference)
            for attr, attr_reference in base_arguments.items()
        }

        # Create child class from this instance.
        control_center = ExpamClassifierProcesses(logging_dir=logging_dir, **base_attributes)

        control_center.set_methods(
            {
                "classify": {
                    "processor": None,
                    "transition": None
                }
            }
        )

        return control_center


class Distribution:
    def __init__(self, k, kmer_db, index, lca_matrix, url, out_dir, temp_dir, logging_dir, alpha,
                 keep_zeros=False, cutoff=0.0, cpm=0.0, paired_end=False):
        """

        :param k: Int
        :param kmer_db: SharedDB
        :param lca_matrix: np.ndarray
        :param url: str
        :param out_dir: str
        """
        # Set up processing references.
        self.k = k
        self.alpha = alpha
        self.kmer_db = kmer_db
        self.lca_matrix = lca_matrix
        self.out_dir = out_dir
        self.temp_dir = temp_dir
        self.logging_dir = logging_dir

        self.index = index
        self.node_names = [node.name if i > 0 else "unclassified" for i, node in enumerate(index.pool)]

        self.keep_zeros = keep_zeros
        self.cutoff = cutoff
        self.cpm = cpm

        # Url containing reads.
        self.url = url
        self.paired_end = paired_end

    def run(self, n):
        """
        Begin processing reads in self.url.
        We have some collection of processes accessing reads from storage
        and then another collection of processes that analyse the reads.

        """
        file_queue = self.prepare_queue(self.paired_end)
        n_classifiers = floor(UNION_RATIO * n)
        n_readers = n - n_classifiers

        mp_config = {
            "name": "expam",
            "phases": ["classify"],
            "layers": {
                "class": ReadWorker,
                "class_args": {
                    "paired_mode": self.paired_end,
                    "n_classifiers": n_classifiers,
                    "logging_dir": self.logging_dir,
                },
                "parent": ControlCenter,
                "child": {
                    "class": ClassifyWorker,
                    "class_args": {
                        "k": self.k,
                        "kmer_db": self.kmer_db,
                        "lca_matrix": self.lca_matrix,
                        "out_dir": join(self.out_dir, PHY_RESULTS),
                        "temp_dir": self.temp_dir,
                        "n_readers": n_readers,
                        "logging_dir": self.logging_dir,
                        "alpha": self.alpha
                    },
                    "parent": ReadWorker,
                    "child": None,
                    "n": n_classifiers
                },
                "n": n_readers
            },
            "timeout": TIMEOUT
        }

        process_network = ExpamClassifierProcesses.from_method_dict(self.logging_dir, mp_config)

        # Load jobs.
        for file_dir in file_queue:
            process_network.load_job(phase="classify", job=file_dir)

        process_network.run()

        # Combine raw read output.
        temporary_files = ls(self.temp_dir, ext=".reads_%s" % EXPAM_TEMP_EXT)
        base_file_names = {
            re.match(r'(\S+)_\d+.reads_%s' % EXPAM_TEMP_EXT, file_name).group(1)
            for file_name in temporary_files
        }

        result_files = []
        for base_file in base_file_names:
            results_file_name = os.path.join(self.out_dir, PHY_RESULTS, RAW_RESULTS, base_file + ".csv")
            results = [
                self.get_data(join(self.temp_dir, file))
                for file in temporary_files
                if file[:len(base_file)] == base_file
            ]

            with open(results_file_name, "w") as f:
                f.write("\n".join(results))

            result_files.append((results_file_name, base_file))

        # Count reads at each node, for each sample.
        counts_dicts = {
            "C": {
                "out_dir": os.path.join(self.out_dir, PHY_RESULTS, "classified_counts.csv"),
                "counts": {}
            },
            "S": {
                "out_dir": os.path.join(self.out_dir, PHY_RESULTS, "splits_counts.csv"),
                "counts": {}
            }
        }
        counts_dicts["U"] = counts_dicts["C"]

        for result_file, sample_name in result_files:
            counts_dicts["C"]["counts"][sample_name] = [0 for _ in range(len(self.node_names))]
            counts_dicts["S"]["counts"][sample_name] = [0 for _ in range(len(self.node_names))]

            with open(result_file, "r") as f:
                for line in f:
                    data = line.split("\t")
                    counts_dicts[data[0]]["counts"][sample_name][int(data[2].lstrip("p"))] += 1

        # Get non-zero rows.
        non_zero_indices = list({
            i
            for _, sample_name in result_files
            for result_type in ("C", "S")
            for i, count in enumerate(counts_dicts[result_type]["counts"][sample_name])
            if (count > 0) or (i == 0)
        })

        def get_cumulative(lst):
            cumulative_counts = lst.copy()

            for i in range(len(self.node_names) - 1, 0, -1):
                cumulative_counts[i] += sum([
                    cumulative_counts[self.index._pointers[child]]
                    for child in self.index.pool[i].children
                ])

            return cumulative_counts

        # Sample-wise summary of counts in kraken-like format.
        summary_format = "%f%%\t%d\t%d"

        # Format phylogenetic node string.
        self.node_names = [
            "p" + self.node_names[i]
            if self.index.pool[i].type == "Branch" and i > 0
            else self.node_names[i]
            for i in range(len(self.node_names))
        ]

        for _, sample_name in result_files:
            total_reads = sum(counts_dicts["C"]["counts"][sample_name]) + sum(counts_dicts["S"]["counts"][sample_name])

            cumulative_classifications = get_cumulative(counts_dicts["C"]["counts"][sample_name])
            cumulative_splits = get_cumulative(counts_dicts["S"]["counts"][sample_name])

            # Employ cutoff.
            if self.cutoff > 0 or self.cpm > 0:
                cutoff = max((total_reads / 1e6) * self.cpm, self.cutoff)
                valid_inds = list({
                    i
                    for cumulative_count in (cumulative_classifications, cumulative_splits)
                    for i, count in enumerate(cumulative_count)
                    if (count >= cutoff) or (i == 0)
                })
                valid_inds.sort()

            else:
                valid_inds = list(range(len(self.node_names)))

            sample_summary = ""
            for i in valid_inds:
                if i not in non_zero_indices and not self.keep_zeros:
                    continue

                if i == 0:
                    line_summary = "unclassified\t" + summary_format % (
                        round(100 * cumulative_classifications[0] / total_reads, 3),
                        cumulative_classifications[0],
                        counts_dicts["C"]["counts"][sample_name][0]) + ("\t" * 3) + "\n"
                else:
                    line_classified_summary = self.node_names[i] + "\t" + summary_format % (
                        round(100 * cumulative_classifications[i] / total_reads, 3),
                        cumulative_classifications[i],
                        counts_dicts["C"]["counts"][sample_name][i])
                    line_splits_summary = summary_format % (
                        round(100 * cumulative_splits[i] / total_reads, 3),
                        cumulative_splits[i],
                        counts_dicts["S"]["counts"][sample_name][i])
                    line_summary = line_classified_summary + "\t" + line_splits_summary + "\n"

                sample_summary += line_summary

            with open(os.path.join(self.out_dir, PHY_RESULTS, sample_name + ".csv"), "w") as f:
                f.write(sample_summary)

        # Save raw counts to disk.
        do_unclassified = True
        for counts_dict in (counts_dicts["C"], counts_dicts["S"]):
            out_dir, counts = counts_dict["out_dir"], counts_dict["counts"]
            df = pd.DataFrame.from_dict(counts)

            if not self.keep_zeros:
                df = df.iloc[non_zero_indices]
                index = [self.node_names[i] for i in non_zero_indices]
            else:
                index = self.node_names

            if "unclassified" in index and not do_unclassified:
                df.drop(0, inplace=True)
                index = index[1:]

            df.index = index
            df.to_csv(out_dir, sep="\t")

            do_unclassified = False

    @staticmethod
    def get_data(file_url):
        with open(file_url, "r") as f:
            data = f.read().strip()

        # Delete temporary file.
        os.remove(file_url)

        return data

    @staticmethod
    def best_match(file_name, file_list):
        def string_intersect(str_one, str_two):
            l = min(len(str_one), len(str_two))

            for i in range(l):
                if str_one[i] != str_two[i]:
                    return str_one[:i]
            else:
                return str_one[:l]

        match_ind = 0
        best_match = string_intersect(file_name, file_list[0])

        for i in range(1, len(file_list)):
            current_match = string_intersect(file_name, file_list[i])

            if len(current_match) > len(best_match):
                best_match = current_match
                match_ind = i

        return match_ind

    def prepare_queue(self, paired_end=False):
        # Create a list of all the files containing reads to be processed.
        file_list = [
            file_name
            for file_name in ls(self.url)
            if re.match(r"(\S+)\.(?:fa|fna|ffn|fq|fasta|fastq)(?:\.tar)*(?:\.gz)*$", file_name)
        ]

        file_queue = []

        if paired_end:
            while file_list:
                next_file = file_list.pop()

                # Find buddy.
                match_ind = self.best_match(next_file, file_list)
                paired_file = file_list.pop(match_ind)

                file_queue.append((join(self.url, next_file), join(self.url, paired_file)))

        else:
            for file_name in file_list:
                file_queue.append((join(self.url, file_name),))

        return file_queue

    @staticmethod
    def _initialise_results(names, unclassified=True):
        """
        Results are returned in spectral form:
            coord --> # hits at coord.

        :param names: List - node names in the tree.
        :return: pd.Series()

        """
        index = (unclassified * ["unclassified"]) + names
        return pd.Series(
            data=np.zeros(len(names) + unclassified, dtype=int),
            index=index)


class ClassificationResults:
    def __init__(self, index, phylogeny_index, in_dir, out_dir, groups=None, keep_zeros=False, cutoff=0.0, cpm=0.0,
                 use_node_names=True, phyla=False, name_taxa=None, colour_list=None, circle_scale=1.0):
        self.index = index
        self.phylogeny_index = phylogeny_index

        self.in_dir = in_dir
        self.out_dir = out_dir

        self.groups = groups  # [(colour, (name1, name2, ...)), (colour, (name3, ...)), ...]
        self.keep_zeros = keep_zeros

        self.cutoff = cutoff  # Cutoff by counts.
        self.cpm = cpm  # Cutoff by counts per million.
        self.use_node_names = use_node_names

        self.phyla = phyla
        self.colour_list = colour_list
        self.name_taxa = name_taxa
        self.circle_scale = circle_scale

        if self.phyla and self.name_taxa is None:
            raise Exception("Inclusion of phyla requires mapping of names to taxonomic lineages!")

        self.tax_id_hierarchy = {"1": set()}  # Map from tax_id -> immediate children.
        self.tax_id_pool = ["1"]  # Children must appear later than parent this list.

    def to_taxonomy(self, name_to_lineage, taxon_to_rank, tax_dir):
        col_names = ["c_perc", "c_cumul", "c_count", "s_perc", "s_cumul", "s_count", "rank", "scientific name"]

        raw_counts_dir = os.path.join(self.in_dir, 'raw')
        raw_output_dir = os.path.join(tax_dir, 'raw')

        if not os.path.exists(raw_output_dir):
            os.mkdir(raw_output_dir)

        class_counts = pd.read_csv(os.path.join(self.in_dir, "classified_counts.csv"), sep="\t", index_col=0, header=0)
        split_counts = pd.read_csv(os.path.join(self.in_dir, "splits_counts.csv"), sep="\t", index_col=0, header=0)

        # Get rid of phylogenetically printed node names.
        def fix_index(df):
            df.index = [
                index.lstrip("p")
                if index not in self.phylogeny_index._pointers
                else index
                for index in df.index
            ]

        fix_index(class_counts)
        fix_index(split_counts)

        samples = class_counts.columns.tolist()
        phylogeny_nodes = split_counts.index.tolist()

        # Map phylogeny index to taxonomy.
        phylogeny_to_taxonomy, tax_id_data = {}, {}
        for phylogeny_node in phylogeny_nodes:
            tax_id, tax_id_map = self.map_phylogeny_node(phylogeny_node, name_to_lineage, taxon_to_rank)

            phylogeny_to_taxonomy[phylogeny_node] = tax_id
            tax_id_data.update(tax_id_map)

        # Get relevant taxon ids.
        tax_ids, ranks, lineages = ["1"], ["root"], [""]
        for tax_id in self.tax_id_pool[1:]:
            rank, lineage = tax_id_data[tax_id]

            tax_ids.append(tax_id)
            ranks.append(rank)
            lineages.append(lineage)

        for sample_name in samples:
            #
            # Map summary output to taxonomy.
            #

            df = pd.DataFrame(0, index=["unclassified"] + tax_ids, columns=col_names)
            total_counts = sum(class_counts[sample_name]) + sum(split_counts[sample_name])

            df.loc[tax_ids[0]:, "rank"] = ranks
            df.loc[tax_ids[0]:, "scientific name"] = lineages

            # Manually insert unclassified counts.
            df.loc["unclassified", "c_count"] = class_counts.loc["unclassified", sample_name]
            df.loc["unclassified", "c_cumul"] = df.loc["unclassified", "c_count"]
            df.loc["unclassified", "c_perc"] = round(100 * df.loc["unclassified", "c_cumul"] / total_counts, 3)

            for col_prefix, counts_df in (("c_", class_counts), ("s_", split_counts)):
                for node in counts_df.index[1:]:
                    df.loc[phylogeny_to_taxonomy[node], col_prefix + "count"] += counts_df.loc[node, sample_name]

            # Get cumulative counts.
            for tax_id in self.tax_id_pool[::-1]:
                for pre in ("c_", "s_"):
                    df.loc[tax_id, pre + "cumul"] = df.loc[tax_id, pre + "count"] + sum([
                        df.loc[child, pre + "cumul"]
                        for child in self.tax_id_hierarchy[tax_id]
                    ])
                    df.loc[tax_id, pre + "perc"] = 100 * df.loc[tax_id, pre + "cumul"] / total_counts

            # Format percentage signs.
            df.loc[:, "c_perc"] = round(df["c_perc"], 3).map(str) + "%"
            df.loc[:, "s_perc"] = round(df["s_perc"], 3).map(str) + "%"

            # Employ cutoff.
            cutoff = max((self.cutoff, (total_counts / 1e6) * self.cpm))
            df = df[(df['c_cumul'] > cutoff) | (df['s_cumul'] > cutoff) | (df.index == 'unclassified')]

            df.to_csv(os.path.join(tax_dir, sample_name + ".csv"), sep="\t", header=True)

            #
            # Map raw read output to taxonomy.
            #

            raw_counts_file = os.path.join(raw_counts_dir, sample_name + ".csv")
            raw_read_data = []

            with open(raw_counts_file, 'r') as f:
                for line in f:
                    read_data = line.strip().split('\t')[:-1]

                    if read_data[0] == 'U':
                        read_data[2] = 'NA'
                    else:
                        phy_id = int(read_data[2].lstrip('p'))
                        phylogeny_node = self.phylogeny_index.pool[phy_id].name
                        read_data[2] = phylogeny_to_taxonomy[phylogeny_node]

                    raw_read_data.append('\t'.join(read_data))

            with open(os.path.join(raw_output_dir, sample_name + '.csv'), 'w') as f:
                f.write('\n'.join(raw_read_data))

    def map_phylogeny_node(self, node_name, name_to_lineage, taxon_to_rank):
        # Map from (node_name) --> (taxid, rank, taxonomic lineage)
        #
        leaves = self.get_leaves(node_name)
        lineages = [name_to_lineage[leaf] for leaf in leaves]
        common_lineage = self.tuple_intersect(*lineages)

        if not common_lineage:
            phy_tax_id, phy_rank = "1", "root"
        else:
            phy_tax_id, phy_rank = taxon_to_rank[common_lineage[-1]]

        tax_id_map = {}
        pool_index = 1
        for i in range(len(common_lineage)):
            parent_tax_id, parent_rank = taxon_to_rank[common_lineage[i]]
            tax_id_map[parent_tax_id] = [parent_rank, " ".join(common_lineage[:i + 1])]

            if i == 0:
                # Connect to root.
                self.tax_id_hierarchy["1"].add(parent_tax_id)

            # Set tax_id in pool.
            try:
                pool_index = self.tax_id_pool.index(parent_tax_id) + 1
            except ValueError:
                self.tax_id_pool.insert(pool_index, parent_tax_id)
                pool_index += 1

            # Set taxa in hierarchy.
            if i + 1 < len(common_lineage):
                child_tax_id, child_rank = taxon_to_rank[common_lineage[i + 1]]
                if parent_tax_id in self.tax_id_hierarchy:
                    self.tax_id_hierarchy[parent_tax_id].add(child_tax_id)
                else:
                    self.tax_id_hierarchy[parent_tax_id] = {child_tax_id}
            elif parent_tax_id not in self.tax_id_hierarchy:
                self.tax_id_hierarchy[parent_tax_id] = set()

        return phy_tax_id, tax_id_map

    def get_leaves(self, node_name):
        return self.phylogeny_index.get_child_leaves(node_name)

    @staticmethod
    def tuple_intersect(*tuples):
        intersection = ()

        length = min([len(tup) for tup in tuples])
        for i in range(length):
            item = tuples[0][i]

            for j in range(1, len(tuples)):
                if tuples[j][i] != item:
                    return intersection

            else:
                intersection += (item,)

        return intersection

    def draw_results(self):
        # Draw classified tree.
        self.phylogeny_index.draw_results(
            os.path.join(self.in_dir, "classified_counts.csv"),
            os.path.join(self.out_dir, "phylotree_classified.pdf"),
            skiprows=[1],
            groups=self.groups,
            cutoff=self.cutoff,
            cpm=self.cpm,
            colour_list=self.colour_list,
            name_to_taxon=self.name_taxa,
            use_phyla=self.phyla,
            keep_zeros=self.keep_zeros,
            use_node_names=self.use_node_names
        )

        # Draw unclassified tree.
        self.phylogeny_index.draw_results(
            os.path.join(self.in_dir, "splits_counts.csv"),
            os.path.join(self.out_dir, "phylotree_splits.pdf"),
            groups=self.groups,
            cutoff=self.cutoff,
            cpm=self.cpm,
            colour_list=self.colour_list,
            name_to_taxon=self.name_taxa,
            use_phyla=self.phyla,
            keep_zeros=self.keep_zeros,
            use_node_names=self.use_node_names
        )



import os
from setuptools import setup, find_packages
from setuptools.extension import Extension

from Cython.Build import cythonize
import numpy as np

EXPAM_VERSION = (0, 0, 8)

SOURCE = os.path.dirname(os.path.abspath(__file__))

# Get project description.
with open(os.path.join(SOURCE, "README.md"), mode="r", encoding="utf-8") as f:
    long_description = f.read()

# Extension instances for Cython scripts.
extensions = [
    Extension(
        "map",
        sources=["src/expam/c/map.pyx"],
        include_dirs=[np.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
    ),
    Extension(
        "extract",
        sources=["src/expam/c/extract.pyx", "src/expam/c/kmers.c", "src/expam/c/jellyfish.c"],
        include_dirs=[np.get_include()],
        extra_compile_args=["-std=c99"],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
    ),
    Extension(
        "sets",
        sources=["src/expam/c/sets.pyx", "src/expam/c/mfil.c"],
        include_dirs=[np.get_include()],
        extra_compile_args=["-std=c99"],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]
    ),
]

setup(
    #
    # Metadata.
    #
    name="expam",
    version="%d.%d.%d" % EXPAM_VERSION,
    description="Metagenomic profiling using a reference phylogeny",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/seansolari/expam",
    author="Sean Solari",
    author_email="sean.solari@monash.edu",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    #
    # Source files.
    #
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    #
    # External dependencies.
    #
    python_requires=">=3.8",
    install_requires=[
        # "ete3",
        # "PyQt5==5.12",
        "numpy",
        "matplotlib",
        "pandas",
        "psutil",
        "requests",
        # "sourmash",
        "tables"
    ],
    #
    # Cython modules.
    #
    ext_package="expam.c",
    ext_modules=cythonize(extensions, language_level="3"),
    #
    # Make main callable from console.
    #
    entry_points={
        "console_scripts": [
            "expam=expam.cli:main",
            "expamlimit=expam.sandbox:main"
        ],
    },
    #
    # Source links.
    #
    project_urls={
        "Bug Reports": "https://github.com/seansolari/expam/issues",
    },
)



import io
import math
import queue
import time
import multiprocessing.shared_memory
import traceback
from multiprocessing import Process, Queue, shared_memory
import os
import re

import numpy as np
import psutil
from expam.c.extract import get_raw_kmers, remove_duplicates, to_32bit, import_mask, map_kmers, \
    bytes_intersect
from expam.c.map import binarysearch, binarysearch_put, classify
from expam.c.sets import filler, disjoint as make_disjoint
from expam.run import log_timer, new_logger
from expam.run import CHUNK_SIZE, EXPAM_TEMP_EXT, KEYS_DATA_TYPE, KEYS_DATA_TYPE_SIZE, \
    MAP_DATA_TYPE, MAP_DATA_TYPE_SIZE, NULL_VALUE, TIMEOUT
from expam.sequences import format_name, get_opener
from expam.stores import MaskStore, ResizableSharedMemory, SequenceStore, TablesDb

COMMAND_CODES = {
    "TRANSITION": 2,
    "ACTIVE": 1,
    "SLEEP": 0,
    "DEAD": -1,
    "ERROR": -2
}


class JobWorker(Process):

    def __init__(self, logging_dir, status_index, statuses,
                 child_statuses, timeout=TIMEOUT, job_name="Worker",
                 phases=None):

        super(JobWorker, self).__init__()

        self._process = None

        self._status_id = status_index
        self._fellow_statuses = statuses
        self._child_statuses = child_statuses

        self.phases = [] if phases is None else list(phases)
        self.work_queue = None  # Where work comes from.
        self.child_queues = None  # Where to distribute work.

        #
        # Defined at ControlCenter.
        self._parents = []
        self._children = []
        self._command_permissions = {
            "command": "None",
            "counts": 0
        }

        #
        # User defined methods.
        self._getters = {}
        self._processors = {}
        self._transitions = {}
        self._shutdown = self.default_shutdown

        self._timeout = timeout
        self.job_name = job_name

        # Logging to disk.
        self.logging_dir = logging_dir
        self.logger = None

    def set_method(self, method_name, phase, method):
        methods = {
            "getter": self._getters,
            "processor": self._processors,
            "transition": self._transitions,
        }

        if method_name not in methods:
            raise ValueError("Invalid method name %s!" % method_name)

        if method is not None:
            methods[method_name][phase] = method

    def init_methods_from_dict(self, config):

        if not isinstance(config, dict):
            raise TypeError("Config must be dictionary not %s!" % str(type(config)))

        for key, value in config.items():
            if value is None:
                continue

            if key == "shutdown":

                if not callable(value):
                    raise TypeError("Shutdown object must be callable!")

                self.set_shutdown(value)

            elif key in self.phases:

                if not isinstance(value, dict):
                    raise TypeError("Method declaration must be dictionary not %s!"
                                    % str(type(value)))

                for method_name, method in value.items():
                    self.set_method(
                        method_name=method_name,
                        phase=key,
                        method=method
                    )

            else:
                raise ValueError("Unknown key in class config: %s!" % key)

    def set_shutdown(self, method):
        self._shutdown = method

    def children_have_obeyed(self, *codes):
        if self._child_statuses is None:
            return True

        return all([
            child_status in codes
            for child_status in self._child_statuses
        ])

    @property
    def is_primary_child(self):
        return True if self._status_id == 0 else False

    @property
    def children_have_errors(self):
        if self._child_statuses is None:
            return False

        return any([
            child_status == COMMAND_CODES["ERROR"]
            for child_status in self._child_statuses
        ])

    @property
    def error_propagated(self):
        if self._child_statuses is None:
            return True

        return all([
            child_status == COMMAND_CODES["ERROR"]
            for child_status in self._child_statuses
        ])

    @property
    def children_dead(self):
        return self.children_have_obeyed(COMMAND_CODES["DEAD"])

    def peers_are_okay(self, cid):
        return all([
            peer_status != COMMAND_CODES["ERROR"]
            for i, peer_status in enumerate(self._fellow_statuses)
            if i != cid
        ])

    def wake_children(self):
        for cid in self.yield_children_id():
            self._child_statuses[cid] = COMMAND_CODES["ACTIVE"]

    @property
    def child_id(self):
        return self._status_id

    @property
    def n_parents(self):
        return len(self._parents)

    @property
    def n_children(self):
        return len(self._children)

    @property
    def n_peers(self):
        return len(self._fellow_statuses)

    @property
    def phase(self):
        return self.phases[0]

    @property
    def status(self):
        return self._fellow_statuses[self._status_id]

    @property
    def name(self):
        return self.job_name

    @status.setter
    def status(self, value):
        assert value in COMMAND_CODES.values()

        self._fellow_statuses[self._status_id] = value

    @property
    def current_command(self):
        return self._command_permissions["command"]

    @current_command.setter
    def current_command(self, value):
        self._command_permissions["command"] = value

    @property
    def command_permissions(self):
        return self._command_permissions["counts"]

    @command_permissions.setter
    def command_permissions(self, value):
        self._command_permissions["counts"] = value

    @property
    def is_supposed_to_die(self):
        return self.status == COMMAND_CODES["ERROR"]

    @property
    def is_supposed_to_sleep(self):
        return self.status == COMMAND_CODES["SLEEP"]

    def reset_command_permissions(self):
        self.current_command = "None"
        self.command_permissions = 0

    def tell_children(self, *args):
        if len(args) == 1:
            args = args * self.n_children

        assert len(args) == self.n_children

        for i in range(self.n_children):
            self.tell_child(i, args[i])

    def tell_child(self, cid, data):

        if not self.child_is_dead(cid):
            self.child_queues[cid].put(data)

    def message_child(self, cid, err_data):
        self._children[cid].send(err_data)

    def yield_children_id(self):
        for i in range(len(self._children)):
            yield i

    def timeout(self):
        time.sleep(self._timeout)

    def is_active(self):
        if self.status == COMMAND_CODES["ACTIVE"]:
            return True
        return False

    def _increment_command(self, command_string):

        if self.current_command == "None":
            self.current_command = command_string

        elif self.current_command != command_string:
            raise ValueError("New command %s interrupted old command flow (%s)!"
                             % (command_string, self.current_command))

        # Check completion condition.
        self.command_permissions += 1

        if self.command_permissions == self.n_parents:

            command_code = COMMAND_CODES[command_string]
            death_code = COMMAND_CODES["DEAD"]

            # Broadcast sleep command to all children.
            self.log("%d moving command to children..." % self.pid)
            self.tell_children((command_code, command_string))
            self.log("%d children told!" % self.pid)

            # Children must have either slept or have died due to no more phases
            while not self.children_have_obeyed(command_code, death_code):
                if self.is_primary_child:
                    has_errors = self.check_children()  # Monitor for errors.

                    if has_errors:
                        return

                # Listen for messages.
                messages = self.check_for_messages()
                if messages:
                    self.begin_error_sequence(messages)

                    return

                self.timeout()

            self.log("%d my children have obeyed!" % self.pid)

            # Start transition to sleep state.
            self.reset_command_permissions()
            self.status = COMMAND_CODES["TRANSITION"]

    def activate(self):
        if not self.is_active():
            self.status = COMMAND_CODES["ACTIVE"]

    def process(self, *args):
        memory_info = self._process.memory_info()
        self.log("%d currently using (RSS: %fGb, VSZ: %fGb) of RAM. Processing data..."
                 % (self.pid,
                    memory_info.rss / 1024 ** 3,
                    memory_info.vms / 1024 ** 3))

        return self._processors[self.phase](*args)

    def sleep(self, *args):
        self.log("%d received sleep command." % self.pid)

        self._increment_command("SLEEP")

        if self.status == COMMAND_CODES["TRANSITION"]:
            self.transition(*args)

            # Sleep when phase transition is complete.
            self.status = COMMAND_CODES["SLEEP"]

            # Wait to be awoken by parent.
            while not self.is_active():
                self.timeout()

            # Wake up children.
            self.wake_children()

    def die(self, *args):
        self.log("%d received die command." % self.pid)

        self._increment_command("DEAD")

        if self.status == COMMAND_CODES["TRANSITION"]:
            self.shutdown(*args)

            # Die when phase transition is complete.
            self.status = COMMAND_CODES["DEAD"]

    def set_error_status(self):
        self.status = COMMAND_CODES["ERROR"]

    def reset_status(self):
        self.status = COMMAND_CODES["ACTIVE"]

    def set_dead_status(self):
        self.status = COMMAND_CODES["DEAD"]

    def die_error(self, *args):
        self.set_error_status()

        # Clear my work queue to shut down safely.
        self.clear_work_queue()

        self.shutdown()

    def clear_work_queue(self):
        while True:
            try:
                _ = self.work_queue.get(block=True, timeout=self._timeout)

            except queue.Empty:
                return

    def get_work(self, *args):

        # First check for any errors raised.
        msg = self.check_for_messages()

        if msg is not False:  # This is an error to be propagated.
            return msg

        else:
            if self.phases and self.phase in self._getters:
                return self._getters[self.phase](*args)

            assert len(args) == 2
            wait, pid = args[0], args[1]

            # Check for more jobs.
            try:
                work = self.work_queue.get(block=True, timeout=self._timeout)

                return work

            except queue.Empty:
                if not wait:
                    return None

    def loop_condition(self, cond):
        if self.received_message():
            return False

        return cond

    def received_message(self):
        if self._parents[0].poll(self._timeout):
            return True

        return False

    def check_for_messages(self):
        if self._parents[0].poll(self._timeout):
            return self._parents[0].recv()

        else:
            return False

    def transition(self, *args):
        if self.phases and self.phase in self._transitions:
            self._transitions[self.phase](*args)

        self.phases.pop(0)
        self.log("%d now at phase %s!" % (self.pid, ("end" if not self.phases else self.phase)))

    def shutdown(self, *args):
        return self._shutdown(*args)

    def default_shutdown(self, *args):
        self.log("%d %s closing..." % (self.pid, self.job_name))

    def raise_error(self, message):
        # If we already have an error message, that takes precedence.
        # Only send if there are no current messages (ie. peers haven't raised).

        self.set_error_status()

        if self._parents[0].poll(self._timeout):
            self.log("%d error was overridden!" % self.pid)

            # Another error was raised.
            self.begin_error_sequence()

        else:
            self._parents[0].send(message)
            self.log("%d error raised..." % self.pid)

            # Await confirmation that my error was received.
            self.log("%d awaiting response..." % self.pid)
            receipt = self._parents[0].recv()
            self.log("%d response received!" % self.pid)

            if receipt is True:
                self.log("%d sent error message!" % self.pid)

                if self.is_primary_child:
                    self.propagate_error(-1, message)

                self.die_error(message[1])

            else:
                self.log("%d error was overridden!" % self.pid)

                # Another error was raised first.
                self.begin_error_sequence(receipt)

    def propagate_error(self, eid, err_data):
        self.log("%d propagating error..." % self.pid)

        # Tell children about error and wait for them to shut down.
        for cid in self.yield_children_id():
            if cid != eid:
                self.message_child(cid, err_data)

        while not self.children_dead:
            self.timeout()

        self.log("%d error propagated!" % self.pid)

    def check_children(self):
        assert self.is_primary_child

        if self.children_have_errors:
            self.log("%d found errors!" % self.pid)
            self.set_error_status()

            # Get error and data.
            for cid in self.yield_children_id():

                if self._children[cid].poll(self._timeout):
                    err_data = self._children[cid].recv()
                    self._children[cid].send(True)  # Send confirmation.

                    break

            else:
                raise Exception("No errors found even though children are dead!")

            self.propagate_error(cid, err_data)

            # Tell parent.
            self.raise_error(err_data)

            return True

        return False

    def child_is_dead(self, cid):
        if self._child_statuses[cid] == COMMAND_CODES["DEAD"]:
            return True

        return False

    def begin_error_sequence(self, err_data=None):
        if err_data is None:
            err_data = self._parents[0].recv()

        if self.is_primary_child:
            self.propagate_error(-1, err_data)

        self.die_error(err_data)

    def log(self, msg):
        self.logger.info(msg)

    def run(self):
        # Activate logging.
        self.logger = new_logger(self.logging_dir, str(self.pid) + "_" + self.job_name)
        self.log("%d %s is now awake!" % (self.pid, self.job_name))

        # Activate memory tracking.
        self._process = psutil.Process(self.pid)

        # Phases loop.
        while not self.is_supposed_to_die and self.phases:

            # Check for errors.
            if self.is_primary_child:
                has_errors = self.check_children()

                if has_errors:
                    break

            # Processing loop.
            for pid in range(self.n_parents):

                try:
                    next_job = self.get_work(False, pid)  # True ==> wait for work.

                    if next_job is None:
                        break

                    command_code, data = next_job

                    if command_code == COMMAND_CODES["ACTIVE"]:
                        self.process(data)

                    elif command_code == COMMAND_CODES["SLEEP"]:
                        self.sleep(data)

                        if self.is_supposed_to_sleep:
                            break  # Check next phase exists.

                    elif command_code == COMMAND_CODES["DEAD"]:
                        self.die(data)

                        if self.is_supposed_to_die:
                            break  # Check next phase exists.

                    elif command_code == COMMAND_CODES["ERROR"]:
                        # Error message is currently propagating.
                        self.begin_error_sequence((command_code, data))

                        break

                    else:
                        raise ValueError("Invalid command code %d!" % command_code)

                except Exception as e:
                    # Log error with context.
                    self.log("*** Exception raised! ***\n" + traceback.format_exc())

                    self.log("Exception encountered in %d! Attempting raise..." % self.pid)

                    err_data = (COMMAND_CODES["ERROR"], e)
                    self.raise_error(err_data)

                    break

        if not self.is_supposed_to_die:
            self.shutdown()

        self.set_dead_status()


class UnionWorker(JobWorker):
    def __init__(self, main_con, lca_matrix, pile_size, logging_dir,
                 lock_value, status_index, statuses, child_statuses,
                 timeout=TIMEOUT, phases=None):

        super(UnionWorker, self).__init__(logging_dir, status_index, statuses, child_statuses,
                                          timeout=timeout, phases=phases)

        # Connection to main process.
        self.job_name = "union"
        self.con = main_con

        # For mapping LCAs in mapping stage.
        self.lca_matrix = lca_matrix

        # Piling parameters.
        self.in_allocations = {}
        self.in_arrays = {}
        self.pile = []

        if pile_size == "inf":
            self.pile_threshold = math.inf

        else:
            pile_size = self.is_int(pile_size)

            if pile_size is not False:
                self.pile_threshold = pile_size

            else:
                self.pile_threshold = 255  # Default.

        # Shared memory allocations.
        self.shm_allocations = {
            "keys": {
                "array": None,
                "allocation": None,
                "shm_name": None,
                "size": None,
                "data_type": KEYS_DATA_TYPE,
                "data_type_size": KEYS_DATA_TYPE_SIZE,
                "active": False,
                "write_offset": None,
            },
            "values": {
                "array": None,
                "allocation": None,
                "shm_name": None,
                "size": None,
                "data_type": MAP_DATA_TYPE,
                "data_type_size": MAP_DATA_TYPE_SIZE,
                "active": False,
                "write_offset": None,
            }
        }

        self.lock_value = lock_value

        # Configure working states.
        self.init_methods_from_dict(
            {
                "import": {
                    "getter": None,
                    "processor": self.import_kmers,
                    "transition": self.finish_import,
                },
                "map": {
                    "getter": None,
                    "processor": self.map,
                    "transition": None,
                },
                "shutdown": self.save_shutdown
            }
        )

    @property
    def shm_keys(self):
        return self.shm_allocations["keys"]["array"]

    @shm_keys.setter
    def shm_keys(self, value):
        self.shm_allocations["keys"]["array"] = value

    @property
    def shm_values(self):
        return self.shm_allocations["values"]["array"]

    @shm_values.setter
    def shm_values(self, value):
        self.shm_allocations["values"]["array"] = value

    @property
    def shm_keys_allocation(self):
        return self.shm_allocations["keys"]["allocation"]

    @shm_keys_allocation.setter
    def shm_keys_allocation(self, value):
        self.shm_allocations["keys"]["allocation"] = value

    @property
    def shm_values_allocation(self):
        return self.shm_allocations["values"]["allocation"]

    @shm_values_allocation.setter
    def shm_values_allocation(self, value):
        self.shm_allocations["values"]["allocation"] = value

    @property
    def shm_keys_name(self):
        return self.shm_allocations["keys"]["shm_name"]

    @shm_keys_name.setter
    def shm_keys_name(self, value):
        self.shm_allocations["keys"]["shm_name"] = value

    @property
    def shm_values_name(self):
        return self.shm_allocations["values"]["shm_name"]

    @shm_values_name.setter
    def shm_values_name(self, value):
        self.shm_allocations["values"]["shm_name"] = value

    @property
    def shm_keys_size(self):
        return self.shm_allocations["keys"]["size"]

    @shm_keys_size.setter
    def shm_keys_size(self, value):
        self.shm_allocations["keys"]["size"] = value

    @property
    def shm_values_size(self):
        return self.shm_allocations["values"]["size"]

    @shm_values_size.setter
    def shm_values_size(self, value):
        self.shm_allocations["values"]["size"] = value

    @property
    def keys_data_type(self):
        return self.shm_allocations["keys"]["data_type"]

    @property
    def keys_data_type_size(self):
        return self.shm_allocations["keys"]["data_type_size"]

    @property
    def values_data_type(self):
        return self.shm_allocations["values"]["data_type"]

    @property
    def values_data_type_size(self):
        return self.shm_allocations["values"]["data_type_size"]

    @property
    def pile_length(self):
        return len(self.pile)

    @property
    def size(self):
        return sum([
            self.product(*kmers.shape, self.keys_data_type_size)
            for kmers in self.yield_kmers()
        ])

    @property
    def shape(self):
        return tuple(
            fn([kmers.shape[i] for kmers in self.yield_kmers()])
            for i, fn in enumerate((sum, max))
        )

    def yield_kmers(self):
        if self.shm_keys is not None:
            yield self.shm_keys

        for arr in self.pile:
            yield arr

    @staticmethod
    def is_int(string):
        try:
            value = float(string)
            int_value = int(value)

            if int_value != value:
                return False

            # Positive definite.
            if int_value < 1:
                return False

            return int_value

        except TypeError:
            return False

    @staticmethod
    def product(*args):

        total = 1
        for arg in args:
            total *= arg

        return total

    def import_kmers(self, data):
        window, shm_data = data
        kmers = self.view_kmers(window, shm_data)

        NO_KMERS_INFO = "No new kmers!"
        self.logger.info("Importing %d kmers at %d."
                         % (kmers.shape[0], self.pid))

        if kmers.shape[0] == 0:

            self.logger.info(NO_KMERS_INFO)

        else:

            # Make disjoint.
            kmers = self.disjoint(kmers)

            # If these kmers have no information to offer, continue.
            if kmers.shape[0] == 0:
                self.logger.info(NO_KMERS_INFO)

            else:
                self.pile.append(kmers)

                # Check if pile needs to be collapsed.
                if self.pile_length >= self.pile_threshold:
                    self.collapse()

                else:
                    self.logger.info("Process %d waiting this time around (%d/%s)."
                                     % (self.pid, self.pile_length, str(self.pile_threshold)))

    def view_kmers(self, window, data):
        start, end = window
        name, shape = data

        # Establish shared_memory connection.
        if name not in self.in_arrays:
            self.in_allocations[name] = multiprocessing.shared_memory.SharedMemory(name)
            self.in_arrays[name] = np.ndarray(
                shape=shape,
                dtype=KEYS_DATA_TYPE,
                buffer=self.in_allocations[name].buf
            )

        return (self.in_arrays[name])[start:end]

    @log_timer
    def disjoint(self, new_kmers):
        # Null case.
        n = new_kmers.shape[0]
        # Establish list of kmers against which to cull.
        comparisons = [self.shm_keys, *self.pile] if self.shm_keys is not None else self.pile

        # Remove any kmers that appear in the pile.
        n = make_disjoint([new_kmers, *comparisons])
        # Copy values into a new array.
        distinct_kmers = np.copy(new_kmers[:n])

        # Demonstrate completion.
        new_kmers[:] = 0

        return distinct_kmers

    def collapse(self):
        # Calculate new array parameters.
        size = self.size
        shape = self.shape

        # Remember old shape for filling.
        old_shape = 0 if self.shm_keys is None else self.shm_keys.shape[0]
        self.logger.info("Process %d sending extension request to shape (%d, %d)."
                         % (self.pid, shape[0], shape[1]))

        # Request extension to memory.
        create = self.shm_keys is None
        self.request_extension(size, create=create)

        # Remap memory extension.
        if self.shm_keys_allocation is None:
            self.shm_keys_allocation = ResizableSharedMemory(self.shm_keys_name)

        else:
            self.shm_keys_allocation.remap(size)

        self.shm_keys = np.ndarray(
            shape=shape,
            dtype=self.keys_data_type,
            buffer=self.shm_keys_allocation.buf
        )

        self.logger.info("Process %d extension successfully mapped." % self.pid)

        # Collapse the pile.
        self.collapse_pile(old_shape)

        self.logger.info("Process %d pile successfully collapsed." % self.pid)

    @log_timer
    def request_extension(self, size, create=False):
        self.con.send(
            (
                "shm_extend",
                (
                    self.shm_keys_name,
                    size,
                    create
                )
            )
        )
        result = self.con.recv()

        # Set shared_memory details.
        if create:
            self.shm_allocations["keys"]["shm_name"] = result
            self.shm_allocations["keys"]["active"] = True

        self.shm_keys_size = size

    @log_timer
    def collapse_pile(self, old_shape):
        size = self.pile_threshold + 1
        list_of_arrays = [arr for arr in self.yield_kmers()]

        # Check arrays are C contiguous.
        for arr in list_of_arrays:
            if not arr.flags["C_CONTIGUOUS"]:
                raise ValueError("Piled array is not C-contiguous!")

        filler(size, old_shape, list_of_arrays)

        # Reset the pile.
        self.pile = []

    def finish_import(self, *args):
        self.check_collapse()
        self.tell_size()

    def check_collapse(self):
        # Check if we need one final collapsing of the pile.
        if self.pile_length > 0:
            self.collapse()

        self.logger.info("Process %d finished." % self.pid)

    def tell_size(self):
        # Check for case where we have received no work.
        if self.shm_keys is not None and not self.is_supposed_to_die:
            self.con.send(
                (
                    "keys_size",
                    self.shape
                )
            )

            values_shm_name = self.con.recv()
            values_shape = self.shape[:1]

            # Initiate values allocation.
            self.shm_values_name = values_shm_name
            self.shm_values_size = self.product(values_shape, self.values_data_type_size)
            self.shm_values_allocation = ResizableSharedMemory(values_shm_name)
            self.shm_values = np.ndarray(
                shape=values_shape,
                dtype=self.values_data_type,
                buffer=self.shm_values_allocation.buf
            )

            self.shm_allocations["values"]["active"] = True

    def map(self, data):
        window, shm_data, node_id = data
        kmers = self.view_kmers(window, shm_data)

        self.logger.info("Process %d mapping %d." % (self.pid, node_id))
        binarysearch_put(
            kmers,
            self.shm_keys,
            self.shm_values,
            self.lca_matrix,
            node_id
        )

        # Demonstrate finished.
        kmers[:] = 0

    def save_shutdown(self, *args):
        self.save_data(*args)
        self.close_stores()
        self.close_shm()

    def save_data(self, *args):

        if self.shm_allocations["keys"]["active"] and self.shm_allocations["values"]["active"]:

            # Find our offset.
            self.con.send(
                (
                    "save",
                    self.shape
                )
            )

            arr_offset, db_path = self.con.recv()

            for arr in ("keys", "values"):
                self.shm_allocations[arr]["write_offset"] = arr_offset

            # Wait until it is my turn to open file.
            while self.loop_condition(self.lock_value.value != arr_offset):
                self.timeout()

            # Save data to disk.
            db = TablesDb(db_url=db_path, edit=True)

            for arr_name in ("keys", "values"):
                writer = db.write_keys if arr_name == "keys" else db.write_values
                arr = self.shm_allocations[arr_name]["array"]

                rows_written = writer(
                    arr=self.shm_allocations[arr_name]["array"],
                    offset=self.shm_allocations[arr_name]["write_offset"]
                )

                if rows_written != arr.shape[0]:
                    self.logger.info("Process %d writing of data failed!" % self.pid)
                else:
                    self.logger.info("Process %d successfully wrote data to %s!" % (self.pid, db_path))

            db.close()

            with self.lock_value.get_lock():
                self.lock_value.value += arr.shape[0]

        # Case where no data has been received.
        else:
            if not self.is_supposed_to_die:
                self.log("Warning! A child has gone this build without any work.")

                self.con.send(
                    (
                        "save",
                        None
                    )
                )

                receipt = self.con.recv()
                self.log(receipt)

    def close_stores(self):
        shm_checks = (
            (
                self.shm_allocations[col]["active"],
                self.shm_allocations[col]["allocation"]
            )
            for col in ("keys", "values")
        )

        for shm_check, shm_all in shm_checks:
            if shm_check:

                # Send an unlink and kill request to the main process.
                try:
                    shm_all.unlink()
                    shm_all.close()

                except FileNotFoundError:
                    pass

        # Log my death.
        self.logger.info("%d closed stores." % self.pid)

    def close_shm(self):
        for allocation in self.in_allocations.values():
            allocation.close()


class ExtractWorker(JobWorker):
    def __init__(self, k, shm_kmer_params, kmer_ranges, node_map, out_dir, logging_dir,
                 status_index, statuses, child_statuses, timeout=TIMEOUT,
                 phases=None):

        super(ExtractWorker, self).__init__(logging_dir, status_index, statuses, child_statuses,
                                            timeout=timeout, phases=phases)

        self.job_name = "extract"
        self.k = k
        self.kmer_ranges = kmer_ranges

        self.chunks = (k // CHUNK_SIZE) + (k % CHUNK_SIZE != 0)
        self.d_type = np.dtype(
            [
                ('index', KEYS_DATA_TYPE),
                *(
                    ('k%d' % chunk, KEYS_DATA_TYPE)
                    for chunk in range(self.chunks)
                )
            ]
        )

        # For mapping LCAs.
        self.node_map = node_map

        # Hold the sequences and masks that I have extracted.
        self.sequence_store = SequenceStore(k, out_dir)
        self.mask_store = MaskStore()

        shm_kmer_name, shm_kmer_shape = shm_kmer_params
        allocation_shape = (shm_kmer_shape[0], shm_kmer_shape[1] + 1)

        # Long-term arrays used for extraction.
        self.allocation = np.zeros(
            shape=allocation_shape,  # Account for index col.
            dtype=KEYS_DATA_TYPE,
            order="C",  # Row-major storage (consecutive k-mer access).
        )

        # Long-term array for sending kmers to children.
        self.out_shape = shm_kmer_shape
        self.out_name = shm_kmer_name
        self.out_allocation = multiprocessing.shared_memory.SharedMemory(shm_kmer_name)
        self.out_array = np.ndarray(
            shape=shm_kmer_shape,
            dtype=KEYS_DATA_TYPE,
            buffer=self.out_allocation.buf
        )

        # Primary queue for getting work during mapping phase.
        self.map_queue = Queue()

        # Initialise methods for execution.
        self.init_methods_from_dict(
            {
                "import": {
                    "getter": None,
                    "processor": self.import_sequence,
                    "transition": self.prepare_map,
                },
                "map": {
                    "getter": self.get_my_sequence,
                    "processor": self.send_for_mapping,
                    "transition": self.empty_queue,
                },
                "shutdown": self.close_shm
            }
        )

    @log_timer
    def import_sequence(self, sequence_url):
        def resize_base(size):
            nonlocal base

            base = base[:size]
            update_view()

        def new_view():
            nonlocal base
            return base.ravel().view(dtype=self.d_type)

        def update_view():
            nonlocal base, view
            view = base.ravel().view(dtype=self.d_type)

        # Collect raw sequence.
        self.log("Importing sequence at %s." % sequence_url)
        seq_id, seq_length = self.sequence_store.add_sequence(sequence_url)

        # Get view on required segment of allocation.
        base = self.allocation[:seq_length]
        view = new_view()

        # Create draft mask and get corresponding kmers.
        new_size = get_raw_kmers(
            sequence=self.sequence_store[seq_id],
            k=self.k,
            arr=base
        )
        resize_base(new_size)

        # Sort by kmers, which also sorts the indices
        view.sort(
            order=['k%d' % chunk for chunk in range(self.chunks)],
            kind="stable"
        )
        self.log("Removing duplicates.")
        new_size = remove_duplicates(base)
        resize_base(new_size)

        # Save mask as 32-bit array.
        self.mask_store[seq_id] = to_32bit(view['index'])
        self.send_kmers(
            ranges=self.kmer_ranges,
            kmers=base[:, 1:]
        )

    @staticmethod
    def clear_temp_arrays(arr):
        arr[:] = 0

    @log_timer
    def send_kmers(self, ranges, kmers, node_id=None):
        target = np.ndarray(shape=(kmers.shape[1]), dtype=kmers.dtype)
        target[:] = ranges[0][0]

        total = 0
        n = len(ranges) - 1
        lower_ind = upper_ind = self.get_index(target, kmers, greater=False)

        # Wait for previous data to be sent.
        while self.loop_condition(not np.all(self.out_array == 0)):
            self.timeout()

        # Place kmers in out_array.
        self.out_array[:kmers.shape[0], :] = kmers[:]

        # Separate data between children.
        for i, (lower_bound, upper_bound) in ranges.items():
            target[:] = upper_bound
            upper_ind = self.get_index(target, kmers)

            # Upper bound can only attain equality for maximum hash value.
            if i == n:
                while (upper_ind < kmers.shape[0]) and self.kmers_equal(target, kmers[upper_ind]):
                    upper_ind += 1

            # Don't send empty work.
            if lower_ind != upper_ind:

                message_data = (
                    (lower_ind, upper_ind),
                    (self.out_name, self.out_shape),
                )
                if node_id is not None:
                    message_data += (node_id,)

                self.tell_child(
                    cid=i,
                    data=(
                        COMMAND_CODES["ACTIVE"],
                        message_data
                    )
                )

                total += upper_ind - lower_ind

            # Beginning next search.
            lower_ind = upper_ind

        self.log("%d / %d kmers sent!" % (total, kmers.shape[0]))

    def get_index(self, target, kmers, greater=True):
        def _check_ind():
            nonlocal ind, kmers

            if (greater and ind == kmers.shape[0]) or ((not greater) and ind == 0):
                return True

            return False

        # Find closest element in list. Note that this element
        # may be less than the bound.
        _, ind = binarysearch(target, kmers)

        if _check_ind():
            return int(ind)

        while (not _check_ind()) and self.kmers_compare(kmers[ind], target, greater):
            if greater:
                ind += 1

            else:
                ind -= 1

        return int(ind)

    @staticmethod
    def kmers_equal(one, two):
        return np.all(one == two)

    @staticmethod
    def kmers_compare(one, two, greater):
        for i in range(one.shape[0]):
            if one[i] < two[i]:
                return greater

            elif one[i] > two[i]:
                return not greater

            elif i + 1 == one.shape[0]:
                return not greater

    def prepare_map(self, *args):
        # Save accession ids from my sequences.
        self.sequence_store.save_accession_ids(suffix=str(self._status_id))

        # Prepare mapping jobs.
        self.load_queue()

    def load_queue(self):
        for seq_id in self.sequence_store.keys():
            self.map_queue.put(
                (
                    COMMAND_CODES["ACTIVE"],
                    seq_id
                )
            )

    def get_my_sequence(self, *args):

        # Try to get local work first.
        try:
            work = self.map_queue.get(block=True, timeout=self._timeout)

            return work

        except queue.Empty:
            pass

        # Now look for commands from parent.
        try:
            work = self.work_queue.get(block=True, timeout=self._timeout)

            return work

        except queue.Empty:
            return None

    def send_for_mapping(self, seq_id):
        node_id = self.node_map[seq_id]

        sequence = self.sequence_store[seq_id]
        mask = self.mask_store[seq_id]
        n_kmers = mask.shape[0]

        # Get view on required segment of allocation.
        base = self.allocation[:n_kmers]
        view = base.ravel().view(dtype=np.dtype(
            [
                ('index', KEYS_DATA_TYPE),
                *(
                    ('k%d' % chunk, KEYS_DATA_TYPE)
                    for chunk in range(self.chunks)
                )
            ]
        ))

        # Place 32-bit mask in 64-bit array.
        import_mask(
            data=mask,
            dest=view['index']
        )
        map_kmers(
            sequence=sequence,
            k=self.k,
            arr=base
        )

        # Send these kmers to be mapped.
        self.send_kmers(
            ranges=self.kmer_ranges,
            kmers=base[:, 1:],
            node_id=node_id
        )

    def empty_queue(self, *args):
        try:
            _ = self.map_queue.get(block=True, timeout=self._timeout)

            raise ValueError("Mapping queue was not empty!")

        except queue.Empty:

            return

    def close_shm(self):
        try:
            self.out_allocation.unlink()
            self.out_allocation.close()

        except FileNotFoundError:
            pass


class ReadWorker(JobWorker):
    def __init__(self, paired_mode, n_classifiers, logging_dir, status_index, statuses, child_statuses,
                 timeout=TIMEOUT, phases=None):
        super(ReadWorker, self).__init__(logging_dir, status_index, statuses, child_statuses, timeout=timeout,
                                         phases=phases)
        self.job_name = "reader"

        self.paired_mode = paired_mode

        self.n_classifiers = n_classifiers
        self._c = 0  # Which classifier do we next send data to?

        self.n_streams = 1 if not paired_mode else 2

        self.init_methods_from_dict(
            {
                "classify": {
                    "getter": None,
                    "processor": self.load_reads,
                    "transition": None,
                }
            }
        )

    def load_reads(self, file_urls):
        print("Loading reads from %s..." % ", ".join(file_urls))

        file_names, file_modes, file_openers, file_types = self.init_openers(*file_urls)

        # Open each file.
        _file_handles = [
            opener(file_url, mode)
            for opener, file_url, mode in zip(file_openers, file_urls, file_modes)
        ]

        read_stream = io.BytesIO()

        read_ids = []
        read_seq = None

        while read_ids or read_seq is None:  # read_seq is None is initial case.
            for i in range(self.n_streams):
                file_type = file_types[i]

                if file_type == "fasta":
                    next_read_id = self.scan_fa_to_next_read(_file_handles[i], read_stream)

                else:  # File type is "fastq".
                    next_read_id = self.scan_fq_to_next_read(_file_handles[i], read_stream)

                if next_read_id is not None:
                    if i == 0:
                        read_ids.append(next_read_id)
                    else:
                        read_ids[-1] = bytes_intersect(read_ids[-1], next_read_id)

            read_seq = read_stream.getvalue()

            if len(read_seq) > self.n_streams:
                read_id = read_ids.pop(0)

                self.send_read(min(file_names), read_id, read_seq)

            read_stream.seek(0)
            read_stream.truncate()

        for i in range(self.n_streams):
            # Close file handles.
            _file_handles[i].close()

        # Close data stream.
        del read_stream

    def send_read(self, file_name, read_id, read):
        self.tell_child(
            cid=self._c,
            data=(
                COMMAND_CODES["ACTIVE"],
                (file_name, read_id, read, self.child_id)
            )
        )

        self._c = (self._c + 1) % self.n_classifiers

    def init_openers(self, *file_urls):
        file_names, file_modes, file_openers, file_types = [], [], [], []

        for file_url in file_urls:
            file_name = format_name(file_url)
            file_name, (mode, opener) = get_opener(file_name)

            if self.check_suffix(file_url, ".fasta", ".fa"):
                file_type = "fasta"

            elif self.check_suffix(file_url, ".fastq", ".fq"):
                file_type = "fastq"

            else:
                raise ValueError("Unknown sequence format for %s!" % file_url)

            file_names.append(file_name)
            file_modes.append(mode)
            file_openers.append(opener)
            file_types.append(file_type)

        return file_names, file_modes, file_openers, file_types

    @staticmethod
    def check_suffix(string, *suffixes):
        for suffix in suffixes:
            if re.findall(r"(%s)(?:.tar)*(?:.gz)*$" % suffix, string):
                return True

        return False

    @staticmethod
    def scan_fa_to_next_read(file, read_stream):
        next_read_id = None

        while True:
            line = file.readline()

            if line == b'':
                read_stream.write(b'N')
                break  # EOF.

            if b'>' in line:
                next_read_id = re.match(b'^>(\S+)\s', line).group(1)

                # End my contribution to this read.
                read_stream.write(b'N')
                break

            else:
                read_stream.write(line.strip())

        return next_read_id

    @staticmethod
    def scan_fq_to_next_read(file, read_stream):
        next_read_id = None

        while True:
            line = file.readline()

            if line == b'':
                read_stream.write(b'N')
                break

            if b'@' in line:
                next_read_id = re.match(b'^@(.*)\s', line).group(1)

                read_stream.write(b'N')
                break

            else:
                read_stream.write(line.strip())

                # Skip next two lines.
                for _ in range(2):
                    file.readline()

        return next_read_id


class ClassifyWorker(JobWorker):
    def __init__(self, k, out_dir, temp_dir, kmer_db, lca_matrix, n_readers, logging_dir, alpha, status_index,
                 statuses, child_statuses, timeout=TIMEOUT, phases=None):
        super(ClassifyWorker, self).__init__(logging_dir, status_index, statuses, child_statuses, timeout=timeout,
                                             phases=phases)
        self.job_name = "classify"
        self.n_readers = n_readers

        # Configuration parameters.
        self.k = k
        self.alpha = alpha
        self.chunks = (k // CHUNK_SIZE) + (k % CHUNK_SIZE != 0)

        self.out_dir = out_dir
        self.temp_dir = temp_dir

        self._read_array_length = 150
        self.read_array = np.ndarray((self._read_array_length, self.chunks + 1), dtype=KEYS_DATA_TYPE)

        self.RAW = "_%s.reads_%s" % ("%d", EXPAM_TEMP_EXT)

        # Load the database.
        self.keys_shm = shared_memory.SharedMemory(
            name=kmer_db.keys_shm_name,
            create=False
        )
        self.keys = np.ndarray(
            shape=kmer_db.keys_meta["shape"],
            dtype=kmer_db.keys_meta["dtype"],
            buffer=self.keys_shm.buf
        )
        self.values_shm = shared_memory.SharedMemory(
            name=kmer_db.values_shm_name,
            create=False
        )
        self.values = np.ndarray(
            shape=kmer_db.values_meta["shape"],
            dtype=kmer_db.values_meta["dtype"],
            buffer=self.values_shm.buf
        )

        self.current_file_data = [("", []) for _ in range(n_readers)]

        # LCA Processing.
        self.lca_matrix = lca_matrix

        self.init_methods_from_dict(
            {
                "classify": {
                    "getter": None,
                    "processor": self.classify_read,
                    "transition": self.save_final_data,
                }
            }
        )

    def classify_read(self, read_data):
        new_file_name, read_id, read, reader_id = read_data

        # Continue to get reads without checking for errors; only check for errors when there is a break in workflow.
        while True:
            read_id = read_id.decode("utf8")

            max_n_kmers = len(read) - self.k + 1
            if max_n_kmers > self.read_array.shape[0]:
                self.read_array = np.ndarray((max_n_kmers, self.chunks + 1), dtype=KEYS_DATA_TYPE)

            if new_file_name != self.current_file_data[reader_id][0]:
                if len(self.current_file_data[reader_id][1]) > 0:
                    self.save_temp_data(reader_id)

                self.current_file_data[reader_id] = (new_file_name, [])

            result, best_clade, read_length, result_string = self.process_read(read)
            self.current_file_data[reader_id][1].append(
                [result, read_id, "p" + str(best_clade), str(read_length), result_string]
            )

            # Get a head start on any available work.
            try:
                command_code, data = self.work_queue.get(block=True, timeout=self._timeout)

                if command_code != COMMAND_CODES["ACTIVE"]:
                    self.work_queue.put((command_code, data))
                    break

                new_file_name, read_id, read, reader_id = data
            except queue.Empty:
                break

    def save_temp_data(self, reader_id):
        def format_result(result_list):
            return "\n".join(["\t".join(item) for item in result_list])

        def base_name(file_url):
            return re.match(r"(\S+)\.(?:fa|fq|fasta|fastq)(?:\.tar)*(?:\.gz)*$", file_url).group(1)

        file_name, file_data = self.current_file_data[reader_id]

        my_file_name = os.path.join(self.temp_dir, base_name(file_name) + (self.RAW % self.pid))
        with open(my_file_name, "a+") as f:
            f.write(format_result(file_data))

    def process_read(self, read):
        read_length = len(read)
        n_kmers = get_raw_kmers(read, self.k, self.read_array)

        # Control for reads that don't have long enough good regions for kmers to be extracted.
        if n_kmers == 0:
            return "U", 0, read_length, ""

        # Map the kmers.
        cls, common_clade, kmer_string = classify(self.read_array[:n_kmers], self.keys, self.values,
                                                  self.lca_matrix, NULL_VALUE, self.alpha)

        if cls == -1:
            result = "U"

        elif cls == 1:
            result = "S"

        else:
            result = "C"

        return result, common_clade, read_length, kmer_string

    def save_final_data(self, *args):
        for reader_id in range(self.n_readers):
            if len(self.current_file_data[reader_id][1]) > 0:
                self.save_temp_data(reader_id)

        self.keys_shm.close()
        self.values_shm.close()
